package jp.veritrans.mpos.v2.sdk.sample.model;

import android.os.Build;
import android.text.Html;
import android.text.SpannableStringBuilder;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import jp.veritrans.mpos.v2.sdk.MposResponseParam;

/**
 * レスポンスパラメータのログ出力用データを生成するクラス。
 */
public final class ResponseLogMaker {

	/**
	 * 並び順に従って比較するComparator。
	 */
	private static class ResponseComparator implements Comparator<String> {
		private Map<String, Integer> priorities = new HashMap<>();

		public ResponseComparator(String[] orders) {
			for (int i = 0; i < orders.length; i++) {
				priorities.put(orders[i], i);
			}
		}

		@Override
		public int compare(String o1, String o2) {
			Integer n1 = priorities.get(o1);
			Integer n2 = priorities.get(o2);
			if (n1 != null && n2 != null) {
				return n1 - n2;
			}
			if (n1 == null && n2 == null) {
				return 0;
			}
			return (n1 != null) ? -1 : 1;
		}

		/**
		 * 初期化時以外で優先順位値を追加
		 * 順位としては末尾
		 *
		 * @param order キー名
		 */
		public void addPriorities(String order) {
			priorities.put(order, priorities.size());
		}
	}

	/**
	 * 指定の並び順に従ってレスポンスをソートして、ログ出力用データに変換して返す。
	 *
	 * @param <T> レスポンスパラメータ値の型
	 * @param params レスポンスパラメータ
	 * @param orders 並び順
	 * @return ソート済みログ出力用データ
	 */
	public static <T> Map<String, ResponseLogItem> convertItems(Map<String, T> params, String[] orders) {
		ResponseComparator comparator = new ResponseComparator(orders);
		TreeMap<String, ResponseLogItem> result = new TreeMap<>(comparator);
		Map<String, T> restMap = new HashMap<>(params);
		String log = (String) restMap.remove("LOG");

		Object status = params.get(MposResponseParam.STATUS);
		boolean success = status != null && status.equals("success");

		// 並び順に存在するものを追加 (必須パラメータ)
		for (String key : orders) {
			Object value = restMap.remove(key);
			if (value != null) {
				ResponseLogItem item = new ResponseLogItem(key, (Serializable) value, ResponseLogItem.Status.OK);
				result.put(key, item);
			} else {
				if (!success) {
					// 失敗時のレスポンスはパラメータ不足チェックをしない
					continue;
				}
				ResponseLogItem item = new ResponseLogItem(key, "", ResponseLogItem.Status.MISSING);
				result.put(key, item);
			}
		}

		// 並び順に存在しないものを追加 (不要パラメータ)
		for (Map.Entry<String, T> entry : restMap.entrySet()) {
			ResponseLogItem item = new ResponseLogItem(entry.getKey(), (Serializable) entry.getValue(), ResponseLogItem.Status.UNUSED);
			String key = entry.getKey();
			// 初期化時に存在しない値のためComparatorに新たに追加する
			comparator.addPriorities(key);
			result.put(key, item);
		}

		// ログを最後尾に追加
		if (log != null) {
			ResponseLogItem item = new ResponseLogItem("LOG", log, ResponseLogItem.Status.OK);
			result.put("LOG", item);
		}

		return result;
	}

	/**
	 * 決済レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertPaymentItems(Map<String, String> params) {
		return convertItems(params, ResponseLogOrder.PAYMENT_ORDERS);
	}

	/**
	 * 取引検索レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertSearchItems(Map<String, Object> params) {
		Map<String, ResponseLogItem> result = convertItems(params, ResponseLogOrder.SEARCH_ORDERS);

		Object historyListObject = params.get(MposResponseParam.HISTORY_LIST);
		if (historyListObject != null) {
			ArrayList<Map<String, ResponseLogItem>> newHistoryList = new ArrayList<>();
			//noinspection unchecked
			List<Map<String, String>> historyList = (List<Map<String, String>>) historyListObject;
			for (Map<String, String> history : historyList) {
				Map<String, ResponseLogItem> items = convertItems(history, ResponseLogOrder.SEARCH_HISTORY_ORDERS);
				newHistoryList.add(items);
			}
			ResponseLogItem item = new ResponseLogItem(MposResponseParam.HISTORY_LIST, newHistoryList, ResponseLogItem.Status.OK);
			result.put(MposResponseParam.HISTORY_LIST, item);
		}

		return result;
	}

	/**
	 * メール送信レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertMailItems(Map<String, Object> params) {
		return convertItems(params, ResponseLogOrder.MAIL_ORDERS);
	}

	/**
	 * 返金レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertRefundItems(Map<String, Object> params) {
		return convertItems(params, ResponseLogOrder.REFUND_ORDERS);
	}

	/**
	 * 取消レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertCancelItems(Map<String, Object> params) {
		return convertItems(params, ResponseLogOrder.CANCEL_ORDERS);
	}

	/**
	 * パスワード変更レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertChangePasswordItems(Map<String, Object> params) {
		return convertItems(params, ResponseLogOrder.CHANGE_PASSWORD_ORDERS);
	}

	/**
	 * 取引集計レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertReportItems(Map<String, Object> params) {
		Map<String, ResponseLogItem> result = convertItems(params, ResponseLogOrder.REPORT_ORDERS);

		Object salesReportObject = params.get(MposResponseParam.SALES_REPORT);
		if (salesReportObject != null) {
			//noinspection unchecked
			Map<String, String> salesReports = (Map<String, String>) salesReportObject;
			TreeMap<String, ResponseLogItem> items = (TreeMap<String, ResponseLogItem>) convertItems(salesReports, ResponseLogOrder.REPORT_SALES_ORDERS);
			ResponseLogItem item = new ResponseLogItem(MposResponseParam.SALES_REPORT, items, ResponseLogItem.Status.OK);
			result.put(MposResponseParam.SALES_REPORT, item);
		}

		Object mileageReportObject = params.get(MposResponseParam.MILEAGE_REPORT);
		if (mileageReportObject != null) {
			//noinspection unchecked
			Map<String, String> mileageReports = (Map<String, String>) mileageReportObject;
			TreeMap<String, ResponseLogItem> items = (TreeMap<String, ResponseLogItem>) convertItems(mileageReports, ResponseLogOrder.REPORT_MILEAGE_ORDERS);
			ResponseLogItem item = new ResponseLogItem(MposResponseParam.MILEAGE_REPORT, items, ResponseLogItem.Status.OK);
			result.put(MposResponseParam.MILEAGE_REPORT, item);
		}

		return result;
	}

	/**
	 * BLEアップデートレスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertBLEUpdateParams(Map<String, Object> params) {
		return convertItems(params, ResponseLogOrder.BLE_UPDATE_ORDERS);
	}

	/**
	 * Tカード会員情報レスポンスパラメータをソートして、ログ出力用データに変換して返す。
	 *
	 * @param params レスポンスパラメータ
	 * @return ソート済みログ出力用データ
	 */
	public static Map<String, ResponseLogItem> convertTCardMemberItems(Map<String, Object> params) {
		return convertItems(params, ResponseLogOrder.T_CARD_ORDERS);
	}

	/**
	 * ログ出力用データを修飾文字列として返す。
	 *
	 * @param items ログ出力用データ
	 * @return 修飾文字列
	 */
	public static CharSequence makeString(Map<String, ResponseLogItem> items) {
		SpannableStringBuilder ssb = new SpannableStringBuilder();

		// ログパラメータは出力しない
		items.remove("LOG");

		// 各パラメータの文字列化
		ssb.append(makeStringFromItems(items));

		return ssb;
	}

	/**
	 * ログ出力用データを再帰的に解析して修飾文字列として返す。
	 * 失敗レスポンスの場合は色分けしない。
	 *
	 * @param items ログ出力用データ
	 * @return 修飾文字列
	 */
	private static CharSequence makeStringFromItems(Map<String, ResponseLogItem> items) {
		StringBuilder sb = new StringBuilder();

		// 不足チェックの有無 (成功時はチェックあり、失敗時はチェックなし)
		Object status = null;
		for (Map.Entry<String, ResponseLogItem> entry : items.entrySet()) {
			if (entry.getKey().equals(MposResponseParam.STATUS)) {
				status = entry.getValue().value;
				break;
			}
		}
		boolean missingCheck = (status != null) && status.equals("success");

		// 並び順に文字列化
		for (Map.Entry<String, ResponseLogItem> entry : items.entrySet()) {
			String key = entry.getKey();
			if (key.equals("LOG")) {
				continue;
			}

			sb.append(makeStringFromElement(entry.getValue(), missingCheck));
		}

		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
			return Html.fromHtml(sb.toString(), Html.FROM_HTML_MODE_LEGACY);
		} else {
			//noinspection deprecation
			return Html.fromHtml(sb.toString());
		}
	}

	/**
	 * ログ出力用データの要素を文字列化する。
	 *
	 * @param obj ログ出力用データの要素
	 * @param missingCheck 不足チェックの有無
	 * @return 文字列
	 */
	private static String makeStringFromElement(Object obj, boolean missingCheck) {
		if (obj == null) {
			return "";
		}

		if (obj instanceof String) {
			return (String) obj;
		}

		if (obj instanceof ResponseLogItem) {
			ResponseLogItem item = (ResponseLogItem) obj;

			String text = "";
			text += item.key;
			text += ": ";
			text += makeStringFromElement(item.value, missingCheck);
			if (!(item.value instanceof List)) {
				text += "<br>";
			}

			switch (item.status) {
				case OK:
					break;
				case MISSING:
					if (!missingCheck) {
						break;
					}
					text = "<font color=\"#A52A2A\">" + text + "</font>";
					break;
				case UNUSED:
					text = "<font color=\"#0000FF\">" + text + "</font>";
					break;
			}

			return text;
		}

		if (obj instanceof List) {
			List<?> list = (List<?>) obj;

			StringBuilder sb = new StringBuilder("[<br>");
			boolean isFirst = true;
			for (Object item : list) {
				if (!isFirst) {
					sb.append(",<br>");
				}

				sb.append(makeStringFromElement(item, missingCheck));

				isFirst = false;
			}
			sb.append("]<br>");
			return sb.toString();
		}

		if (obj instanceof Map) {
			Map<?, ?> map = (Map<?, ?>) obj;

			StringBuilder sb = new StringBuilder("{<br>");
			for (Map.Entry<?, ?> entry : map.entrySet()) {
				sb.append(makeStringFromElement(entry.getValue(), missingCheck));
			}
			sb.append("}<br>");
			return sb.toString();
		}

		return "";
	}
}
